; -------------------------------------
;   Z80 native Virtual Machine 
;   (c) 2015 - 2017 kio@little-bat.de
; -------------------------------------


; Z80 Registers:
;
; AF	used
; BC	used
; DE	VM vstack TOP value (last argument or return value)
; HL	VM vstack SOT value (2nd-last argument or high word of 32 bit value)
; 
; AF' … HL'	unused (available for interrupt)
; 
; IX	
; IY
; SP	stack pointer: z80, vstack, rstack
; PC	program address counter



#local
#code VCC_CORE


; -----------------------------------------
; Helper:
; -----------------------------------------

; ________________________________
; dereference handle hl --> hl
; 4 bytes, 24cc
; mod: a

.macro deref_hl
	ld	a,(hl++)
	ld	h,(hl)
	ld	l,a
.endm

; ________________________________
; dereferene handle de --> de:	
; 4 bytes, 24cc
; mod: hl

.macro deref_de
	ex	de,hl
	ld	e,(hl++)
	ld	d,(hl)
.endm


; -----------------------------------------
; Code Flow:
; -----------------------------------------

.macro RETURN_VOID
	ret
.endm

.macro RETURN_de
	pop	de
	ret
.endm

.macro RETURN_hlde
	pop	hl
	pop	de
	ret
.endm

; drop word or long value
; directly, behind 1, 2 or 3 other values (rval, retaddr):

.macro DROP
	pop	af
.endm

.macro DROP2
	pop	af
	pop	af
.endm
	
.macro NIP
	pop	de
	pop	af
	push	de
.endm

.macro NIP2
	pop	de
	pop	af
	pop	af
	push	de
.endm
	
.macro NIP_OVER2
	pop	hl
	pop	de
	pop	af
	push	de
	push	hl
.endm

.macro NIP2_OVER2
	pop	hl
	pop	de
	pop	af
	pop	af
	push	de
	push	hl
.endm
	
.macro NIP_OVER3
	pop	bc
	pop	hl
	pop	de
	pop	af
	push	de
	push	hl
	push	bc
.endm

.macro NIP2_OVER3
	pop	bc
	pop	hl
	pop	de
	pop	af
	pop	af
	push	de
	push	hl
	push	bc
.endm
	
jp_hl::		jp	hl
jp_iy::		jp	iy


; _______________________________
; call proc ptr
;
; CALLPROCPTR       ( … addr -- [bcde] )
; CALLPROCPTR_IY    ( … addr -- [bcde] )

.macro CALLPROCPTR
 	pop	hl
	call	jp_hl
.endm

.macro CALLPROCPTR_IY
 	pop	iy
	call	jp_iy
.endm




; ________________________________
; jp to label
;
; JP ( [bcdehl] $.w -- [bcdehl] )
; JR ( [bcdehl] $.b -- [bcdehl] )	; $-128 .. $+127

.macro JP $label
 	jp	$label
.endm

.macro JR $label
	jr	$label
.endm

; ________________________________
; jp to label if n=FALSE
;
; JP0  ( n $.w -- )
; AND0 ( n $.w -- )
; JR0  ( n $.b -- )
; AND0 ( n $.b -- )

.macro JP0 $label
	pop	de
	ld	a,d
	or	e
	jp	z,$label
.endm

.macro AND0JP $label
	pop	de
	ld	a,d
	or	e
	push	de
	jp	z,$label
	pop	de
.endm

.macro JR0 $label
	pop	de
	ld	a,d
	or	e
	jr	z,$label
.endm

.macro AND0 $label
	pop	de
	ld	a,d
	or	e
	push	de
	jr	z,$label
	pop	de
.endm
 
; ________________________________
; jp to label if n=TRUE
;
; JP1   ( n $.w -- )
; OR1JP ( n $.w -- )
; JR1   ( n $.b -- )
; OR1   ( n $.b -- )

.macro JP1 $label
	pop	de
	ld	a,d
	or	e
	jp	nz,$label
.endm

.macro OR1JP $label
	pop	de
	ld	a,d
	or	e
	push	de
	jp	nz,$label
	pop	de
.endm

.macro JR1 $label
	pop	de
	ld	a,d
	or	e
	jr	nz,$label
.endm

.macro OR1 $label
	pop	de
	ld	a,d
	or	e
	push	de
	jr	nz,$label
	pop	de
.endm

; ________________________________
; if n<table.count jump to table[n] else resume behind table[]
; table can have from 1 to 256 entries
;
; SWITCH ( n $tablesize.b $default_label.w $table.w[] -- )
;
; invocation:
;
;	call	switch
;	db	tablesize-1
;	dw	default_label
;	dw	labels[tablesize]

.macro SWITCH
	pop	de
	call 	switch
	;db	tablesize-1
	;dw	default_label
	;dw	labels[tablesize]
.endm
	
switch::
	pop	hl		; hl -> tablesize.b
	ld	a,(hl++)	; a = tablesize-1, hl -> default_label

	rlc	d		; test de.hi
	jr	nz,1$		; de.hi != 0 -> oorange

	cp	a,e		; sz-1 - idx
	jr	c,1$		; sz-1<idx -> sz≤idx -> idx≥sz -> beyond table

	; de inside table:
	ex	de,hl		; hl = idx
	inc	hl		; +1 to skip over default entry
	add	hl,hl		; hl = idx*2 for word-size entries
	add	hl,de		; add table base

1$:	ld	a,(hl++)	; get entry: low byte ..
	ld	h,(hl)		; .. high byte
	ld	l,a

	jp	hl		; jump to label






	



#include "abort.s"
#include "page0.s"
#include "core8.s"
#include "core16.s"
#include "core32.s"
#include "core_str.s"
#include "memory.s"

#endlocal


